#include "header\gba.h"
#include "header\bg.h"
#include "header\timer.h"
#include "header\dma.h"
#include "header\irq.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\blend.h"

#include "lib.h"
#include "resource.h"

#include "syscall.h"

//#include "header\font\BGFont.h"
//#include "header\font\BitMapFont.h"

#include <stdlib.h>
#include <math.h>

//Mode0
//SPR LN^

//---------------------------------------------------------------------------
#define STATE_INIT 				1				//is
#define STATE_RESET				2
#define STATE_READY				3
#define STATE_GAME				4

#define MYSHIP_MAX_X			(f32)50			//ړ͈
#define MYSHIP_MAX_Y			(f32)50
#define MYSHIP_VROLL			(f32)1			//[x
#define MYSHIP_MAX_ROLL			(f32)20			//[͈
#define MYSHIP_SQRT2			(f32)0.7f		//΂߈ړ̑x  1/sqrt(2.0f)

#define MYSHIP_BEAM_MIN_POWER	15				//r[̃p[
#define MYSHIP_SHOT_MAX_GAUGE	15				//Vbg̃Q[W
#define MYSHIP_BEAM_MAX_TIME	2				//r[̑ҋ@
#define MYSHIP_SHOT_MAX_TIME	6				//Vbg̑ҋ@

#define BEAM_MAX_CNT			12
#define SHOT_MAX_CNT			30
#define BULLET_MAX_CNT			60
#define SPRITE_MAX_CNT			128

#define PI						(f32)3.141592654f

//---------------------------------------------------------------------------
typedef struct {
	f32 x;
	f32 y;
	s16 roll;			//[
	s16 shotGauge;		//Vbg̃Q[W
	s16 shotTime;		//ҋ@
	s16 beamPower;		//r[̃p[
	s16 beamTime;		//ҋ@
	u8  sprNo;
} ST_MYSHIP;

typedef struct {
	bool sw;
//	f32  x;
	f32  y;
	u8   sprNo;
} ST_BEAM;

typedef struct {
	bool sw;
	f32  x;
	f32  y;
	f32  vx;
	f32  vy;
	u8   sprNo;
} ST_SHOT;

typedef struct {
	bool sw;
	f32  x;
	f32  y;
	f32  vx;
	f32  vy;
	u16  timer;
	void (*pMove)(void*);
	u8   sprNo;
} ST_BULLET;

typedef struct {
	bool sw;
} ST_SPR_MGT;

//---------------------------------------------------------------------------
//lib.c
extern ST_TIMER Timer;

//---------------------------------------------------------------------------
ST_MYSHIP  MyShip;
ST_BEAM    Beam[BEAM_MAX_CNT];
ST_SHOT    Shot[SHOT_MAX_CNT];
ST_BULLET  Bullet[BULLET_MAX_CNT];
ST_SPR_MGT SprMgt[SPRITE_MAX_CNT];

s8         ShotSprCnt;
s8         BeamSprCnt;
s8         BulletSprCnt;

u8         State;						//is󋵃tO
//---------------------------------------------------------------------------
void StateInit(); 
void StateReset();
void StateGame();

void InitWS();
void InitKey();
void IRQUserHandler();
void InitSprite();
void InitSprMgt();

void MoveMain();
void DrawMain();

u8   SprAlloc();
void SprFree(u8 num);

bool CommonOut(f32 x, f32 y, f32 size);
s16  CommonToFieldX(f32 x);
s16  CommonToFieldY(f32 y);

u16  ScanKey();

//----
void MyShipInit();
bool MyShipNew(f32 x, f32 y);
bool MyShipMove();
void MyShipShot();
void MyShipBeam();
void MyShipDeleteBeam();
void MyShipDraw();

//----
void BeamInit();
void BeamNew(f32 y);
void BeamMove();
void BeamDraw();

//----
void ShotInit();
void ShotNew(f32 x, f32 y, f32 dir);
void ShotMove();
void ShotDraw();

//----
void BulletInit();
void BulletNew(f32 x, f32 y, f32 dir, void pMove(void*));
void BulletMove();
void BulletMoveDir(ST_BULLET* p);
void BulletMoveHoming(ST_BULLET* p);
void BulletMoveSplit(ST_BULLET* p);
void BulletDraw();

//---------------------------------------------------------------------------
int main()
{
	State = STATE_INIT;

	for(;;)
	{
		switch(State)
		{
			case STATE_INIT:  StateInit();  break;
			case STATE_RESET: StateReset(); break;
			case STATE_GAME:  StateGame();  break;

			default:
				DebugMessage((u8*)"Err: Main State");
				break;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
void StateInit()
{
	InitWS();
	InitKey();
	InitSprite();

	SetMode(MODE_0 | BG2_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);

	State = STATE_RESET;
}
//---------------------------------------------------------------------------
void InitWS()
{
	REG_WSCNT = CST_ROM0_1ST_3WAIT | CST_ROM0_2ND_1WAIT;
}
//---------------------------------------------------------------------------
void InitKey()
{
	KeyInit();

	REG_INTERUPT = (u32)IRQUserHandler;
	REG_IE       = IRQ_BIT_TIMER0;
	REG_IME      = IRQ_MASTER_ON;
	REG_TM0D     = 65536 - 2621;			// (((16*1024*1024) / 64) * 10) / 1000
	REG_TM0CNT   = TM_ENABLE | TM_FREQ_PER_64 | TM_USEIRQ;
}
//---------------------------------------------------------------------------
void IRQUserHandler()
{
	Timer.clockTick += TIMER_INTMS;
	REG_IF = IRQ_BIT_TIMER0;
}
//---------------------------------------------------------------------------
void InitSprite()
{
	SpriteSetPal16((u16*)&spr_chr_pal_bin);
	SpriteSetData ((u16*)&spr_chr_img_bin, SPR_CHR_IMG_SIZE);
}
//---------------------------------------------------------------------------
void InitSprMgt()
{
	s16 i;

	for(i=0; i<SPRITE_MAX_CNT; i++)
	{
		SprMgt[i].sw = FALSE;

		SpriteMove(i, SCREEN_CX, SCREEN_CY);
	}
}
//---------------------------------------------------------------------------
void StateReset()
{
	InitSprMgt();

	MyShipInit();
	BeamInit();
	ShotInit();
	BulletInit();

	MyShipNew(0, 30);

	State = STATE_GAME;
}
//---------------------------------------------------------------------------
void MyShipInit()
{
	MyShip.x         = 0;
	MyShip.y         = 0;
	MyShip.roll      = 0;
	MyShip.shotGauge = 0;
	MyShip.shotTime  = 0;
	MyShip.beamPower = 0;
	MyShip.beamTime  = 0;
	MyShip.sprNo     = 0;

	SpriteSetSize(MyShip.sprNo, SP_SIZE_16, SP_SQUARE, SP_COLOR_16);
}
//---------------------------------------------------------------------------
void BeamInit()
{
	s16 i;

	for(i=0; i<BEAM_MAX_CNT; i++)
	{
		Beam[i].sw    = FALSE;
		Beam[i].y     = 0;
		Beam[i].sprNo = 0;
	}
	BeamSprCnt = 0;
}
//---------------------------------------------------------------------------
void ShotInit()
{
	s16 i;

	for(i=0; i<SHOT_MAX_CNT; i++)
	{
		Shot[i].sw    = FALSE;
		Shot[i].x     = 0;
		Shot[i].y     = 0;
		Shot[i].vx    = 0;
		Shot[i].vy    = 0;
		Shot[i].sprNo = 0;
	}
	ShotSprCnt = 0;
}
//---------------------------------------------------------------------------
void BulletInit()
{
	s16 i;

	for(i=0; i<BULLET_MAX_CNT; i++)
	{
		Bullet[i].sw    = FALSE;
		Bullet[i].x     = 0;
		Bullet[i].y     = 0;
		Bullet[i].vx    = 0;
		Bullet[i].vy    = 0;
		Bullet[i].timer = 0;
		Bullet[i].pMove = NULL;
		Bullet[i].sprNo = 0;
	}
	BulletSprCnt = 0;
}
//---------------------------------------------------------------------------
bool MyShipNew(f32 x, f32 y)
{
	MyShip.x     = x;
	MyShip.y     = y;
	MyShip.sprNo = 0;

	return TRUE;
}
//---------------------------------------------------------------------------
void StateGame()
{
	DrawMain();
	MoveMain();

	static u16 Cnt  = 0;
	s16 i;

	if(Timer.clockTick % 500 != 0)
	{
		return;
	}

	switch (Cnt++)
	{

	//_e
	case 0:
		BulletNew(0, -40, atan2(MyShip.y + 40, MyShip.x), (void*)&BulletMoveDir);
		break;

	//e
	case 1:
		for(i=0; i<20; i++)
		{
			BulletNew(0, -40, i * 0.314 , (void*)&BulletMoveDir);
		}
		break;

	//ԐtUe
	case 2:
		BulletNew(0, -40, atan2(MyShip.y + 40, MyShip.x), (void*)&BulletMoveHoming);
		break;

	//e
	case 3:
		BulletNew(0, -40, 0.314 * 5, (void*)&BulletMoveSplit);
		break;


	default:
		Cnt = 0;
		break;
	}
}
//---------------------------------------------------------------------------
void DrawMain()
{
	WaitForVsync();

	MyShipDraw();
	BeamDraw();
	ShotDraw();
	BulletDraw();

	//TODO ScoreDraw();
}
//---------------------------------------------------------------------------
void MyShipDraw()
{
	s16 x = CommonToFieldX(MyShip.x) - 8;	//8́AMYSHIP_CX / 2
	s16 y = CommonToFieldY(MyShip.y) - 8;

	SpriteSetChr(MyShip.sprNo, (MyShip.roll/10+3)*2);
	SpriteMove  (MyShip.sprNo, x, y);
}
//---------------------------------------------------------------------------
void BeamDraw()
{
	s16 i;

	for(i=0; i<BEAM_MAX_CNT; i++)
	{
		if(Beam[i].sw == TRUE)
		{
			s16 x = CommonToFieldX(MyShip.x ) - 4;
			s16 y = CommonToFieldY(Beam[i].y) - 4;	//4́ABEAM_CY / 2

			SpriteMove(Beam[i].sprNo, x, y);
		}
	}
}
//---------------------------------------------------------------------------
void ShotDraw()
{
	s16 i;

	for(i=0; i<SHOT_MAX_CNT; i++)
	{
		if(Shot[i].sw == TRUE)
		{
			s16 x = CommonToFieldX(Shot[i].x) - 4;	//4́ASHOT_CX / 2
			s16 y = CommonToFieldY(Shot[i].y) - 4;

			SpriteMove(Shot[i].sprNo, x, y);
		}
	}
}
//---------------------------------------------------------------------------
void BulletDraw()
{
	s16 i;

	for(i=0; i<BULLET_MAX_CNT; i++)
	{
		if(Bullet[i].sw == TRUE)
		{
			s16 x = CommonToFieldX(Bullet[i].x) - 4;	//4́ASHOT_CX / 2
			s16 y = CommonToFieldY(Bullet[i].y) - 4;

			SpriteMove(Bullet[i].sprNo, x, y);
		}
	}
}
//---------------------------------------------------------------------------
void MoveMain()
{
	//TODO ꎞ~

	MyShipMove();
	BeamMove();
	ShotMove();
	BulletMove();
}
//---------------------------------------------------------------------------
bool MyShipMove()
{
	//ړx
	f32 vx = 0.8f;
	f32 vy = 0.8f;

	ST_MYSHIP* p = &MyShip;
	u16 Key      = ScanKey();

	//Vbgƃr[̔˂܂
	if(Key & KEY_A)
	{
		//p[ȂƂAVbg𔭎˂܂
		if(p->beamPower < MYSHIP_BEAM_MIN_POWER)
		{
			p->beamPower++;
			p->shotGauge = MYSHIP_SHOT_MAX_GAUGE;

			if(p->shotTime == 0) MyShipShot();
		} 
		else
		{
			p->shotGauge = 0;
			if(p->beamTime == 0) MyShipBeam();

			vx *= 0.5f;
			vy *= 0.5f;
		}
	}
	else
	{
		//Q[WcĂƂVbg𔭎˂܂
		if(p->shotGauge > 0)
		{
			if(p->shotTime == 0) MyShipShot();
			p->shotGauge--;
		}

		MyShipDeleteBeam();
		p->beamPower = 0;
	}

	//Vbgƃr[̑ҋ@ԂXV
	p->shotTime = (p->shotTime + 1) % MYSHIP_SHOT_MAX_TIME;
	p->beamTime = (p->beamTime + 1) % MYSHIP_BEAM_MAX_TIME;

	//Eړƃ[
	if(Key & KEY_LEFT)
	{
		vx = -vx;
		if(p->roll > -MYSHIP_MAX_ROLL) p->roll -= MYSHIP_VROLL;
	}
	else if(Key & KEY_RIGHT)
	{
		if(p->roll <  MYSHIP_MAX_ROLL) p->roll += MYSHIP_VROLL;
	}
	else
	{
		vx = 0;
		if(!(p->roll == 0))
		{
			p->roll += (p->roll < 0) ? MYSHIP_VROLL : -MYSHIP_VROLL;
		}
	}

	//㉺ړ
	if(Key & KEY_UP)
	{
		vy = -vy;
	}
	else if( !(Key & KEY_DOWN) )
	{
		vy = 0;
	}

	//΂ߕ̑x
	if(vx != 0 && vy != 0)
	{
		vx *= MYSHIP_SQRT2;
		vy *= MYSHIP_SQRT2;
	}

	//ړ
	p->x += vx;
	p->y += vy;
	
	//ړ͈͂̐
	if(p->x < -MYSHIP_MAX_X)      p->x = -MYSHIP_MAX_X;
	else if (MYSHIP_MAX_X < p->x) p->x =  MYSHIP_MAX_X;

	if(p->y < -MYSHIP_MAX_Y)      p->y = -MYSHIP_MAX_Y;
	else if (MYSHIP_MAX_Y < p->y) p->y =  MYSHIP_MAX_Y;

	return TRUE;
}
//---------------------------------------------------------------------------
u16 ScanKey()
{
	if( KeyWaitTimer() == TRUE )
	{
		return KEY_NONE;
	}
	if( KeyRepeat() == TRUE )		//keyGet֐ĂԑOɕKsȂ
	{
		return KEY_NONE;
	}

	return KeyInput();
}
//---------------------------------------------------------------------------
void MyShipShot()
{
	ShotNew(MyShip.x, MyShip.y, 0.70f);
	ShotNew(MyShip.x, MyShip.y, 0.75f);
	ShotNew(MyShip.x, MyShip.y, 0.80f);
}
//---------------------------------------------------------------------------
void ShotNew(f32 x, f32 y, f32 dir)
{
	s16 i;

	for(i=0; i<SHOT_MAX_CNT; i++)
	{
		if(Shot[i].sw == FALSE)
		{
			u8 num = SprAlloc();
			if(num == 0)
			{
				return;
			}

			Shot[i].sw    = TRUE;
			Shot[i].x     = x;
			Shot[i].y     = y;
			Shot[i].vx    = 2.4f * cos(PI * 2.0 * dir);
			Shot[i].vy    = 2.4f * sin(PI * 2.0 * dir);
			Shot[i].sprNo = num;

			SpriteSetSize(Shot[i].sprNo, SP_SIZE_8, SP_SQUARE, SP_COLOR_16);
			SpriteSetChr (Shot[i].sprNo, 66);

			ShotSprCnt++;
			return;
		}
	}
}
//---------------------------------------------------------------------------
void ShotMove()
{
	s16 i;

	for(i=0; i<SHOT_MAX_CNT; i++)
	{
		if(Shot[i].sw == TRUE)
		{
			Shot[i].x += Shot[i].vx;
			Shot[i].y += Shot[i].vy;

			if( CommonOut(Shot[i].x, Shot[i].y, 5) == TRUE )
			{
				Shot[i].sw = FALSE;

				SprFree(Shot[i].sprNo);
				ShotSprCnt--;
			}
		}
	}
}
//---------------------------------------------------------------------------
void MyShipBeam()
{
	BeamNew(MyShip.y);
}
//---------------------------------------------------------------------------
void BeamNew(f32 y)
{
	s16 i;

	for(i=0; i<BEAM_MAX_CNT; i++)
	{
		if(Beam[i].sw == FALSE)
		{
			u8 num = SprAlloc();
			if(num == 0)
			{
				return;
			}

			Beam[i].sw    = TRUE;
			Beam[i].y     = y;
			Beam[i].sprNo = num;

			SpriteSetSize(Beam[i].sprNo, SP_SIZE_8, SP_SQUARE, SP_COLOR_16);
			SpriteSetChr (Beam[i].sprNo, 67);

			BeamSprCnt++;
			return;
		}
	}
}
//---------------------------------------------------------------------------
void MyShipDeleteBeam()
{
	s16 i;

	for(i=0; i<BEAM_MAX_CNT; i++)
	{
		if(Beam[i].sw == TRUE)
		{
			Beam[i].sw = FALSE;

			SprFree(Beam[i].sprNo);
			BeamSprCnt--;
		}
	}
}
//---------------------------------------------------------------------------
void BeamMove()
{
	s16 i;

	for(i=0; i<BEAM_MAX_CNT; i++)
	{
		if(Beam[i].sw == TRUE)
		{
			Beam[i].y -= 6.0f;

			if( CommonOut(MyShip.x, Beam[i].y, 8) == TRUE )
			{
				Beam[i].sw = FALSE;

				SprFree(Beam[i].sprNo);
				BeamSprCnt--;
			}
		}
	}
}
//---------------------------------------------------------------------------
void BulletMove()
{
	s16 i;

	for(i=0; i<BULLET_MAX_CNT; i++)
	{
		if(Bullet[i].sw == TRUE)
		{
			Bullet[i].pMove( &Bullet[i] );

			if( CommonOut(Bullet[i].x, Bullet[i].y, 4) == TRUE )
			{
				Bullet[i].sw = FALSE;

				SprFree(Bullet[i].sprNo);
				BulletSprCnt--;
			}
		}
	}
}
//---------------------------------------------------------------------------
//e + _e
void BulletMoveDir(ST_BULLET* p)
{
	p->x += p->vx;
	p->y += p->vy;
}
//---------------------------------------------------------------------------
//ԐtUe
void BulletMoveHoming(ST_BULLET* p)
{
	p->timer++;

	if(p->timer < 40)
	{	f32 dir = atan2(MyShip.y - p->y, MyShip.x - p->x);
		p->vx   = 1.2f * cos(dir);
		p->vy   = 1.2f * sin(dir);
	}

	p->x += p->vx;
	p->y += p->vy;
}
//---------------------------------------------------------------------------
//e
void BulletMoveSplit(ST_BULLET* p)
{
	p->timer++;

	p->x += p->vx;
	p->y += p->vy;

	if(p->timer > 30)
	{
		BulletNew(p->x, p->y, 0.314 * 4, (void*)&BulletMoveDir);
		BulletNew(p->x, p->y, 0.314 * 6, (void*)&BulletMoveDir);
		BulletNew(p->x, p->y, 0.314 * 5, (void*)&BulletMoveDir);

		p->sw = FALSE;

		SprFree(p->sprNo);
		BulletSprCnt--;
	}
}
//---------------------------------------------------------------------------
void BulletNew(f32 x, f32 y, f32 dir, void pMove(void*))
{
	s16 i;

	for(i=0; i<BULLET_MAX_CNT; i++)
	{
		if(Bullet[i].sw == FALSE)
		{
			u8 num = SprAlloc();
			if(num == 0)
			{
				return;
			}

			Bullet[i].sw    = TRUE;
			Bullet[i].x     = x;
			Bullet[i].y     = y;
			Bullet[i].vx    = 1.2f * cos(dir);
			Bullet[i].vy    = 1.2f * sin(dir);
			Bullet[i].pMove = pMove;
			Bullet[i].timer = 0;
			Bullet[i].sprNo = num;
			
			SpriteSetSize(Bullet[i].sprNo, SP_SIZE_8, SP_SQUARE, SP_COLOR_16);
			SpriteSetChr (Bullet[i].sprNo, 68);

			BulletSprCnt++;
			return;
		}
	}
}
//---------------------------------------------------------------------------
u8 SprAlloc()
{
	s16 i;

	for(i=1; i<SPRITE_MAX_CNT; i++)
	{
		if(SprMgt[i].sw == FALSE)
		{
			SprMgt[i].sw = TRUE;
			return i;
		}
	}

	//Error
	return 0;
}
//---------------------------------------------------------------------------
void SprFree(u8 num)
{
	SprMgt[num].sw = FALSE;
	SpriteMove(num, SCREEN_CX, SCREEN_CY);
}
//---------------------------------------------------------------------------
bool CommonOut(f32 x, f32 y, f32 size)
{
	return x<-50-size || 50+size<x || y<-50-size || 50+size<y;
}
//---------------------------------------------------------------------------
s16 CommonToFieldX(f32 x)
{
	return (s16)((50.0 + x) * 2.4);
}
//---------------------------------------------------------------------------
s16 CommonToFieldY(f32 y)
{
	return (s16)((50.0 + y) * 1.6);
}
